# HEADER
#                     arg/GUI/Logic/argUserSettingsReader.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import os

import yaml
from PySide2.QtCore import QObject

app = "ARG-GUI"

argPythonExePathUserSettings = "python_executable"
argPythonSitePackagePathUserSettings = "python_site_package"
argScriptPathUserSettings = "arg_script"
argLatexProcessorPathUserSettings = "latex_processor"
argSettings = [argPythonExePathUserSettings,
               argPythonSitePackagePathUserSettings,
               argScriptPathUserSettings,
               argLatexProcessorPathUserSettings]


class argUserSettingsReader(QObject):
    """A reader class to read ARG settings file
    """

    def __init__(self, adminLvlPath, usrLvlPath):
        super().__init__()

        self.adminLvlPath = adminLvlPath
        self.usrLvlPath = usrLvlPath
        self.settings = {}

    def read(self):
        """Read specified file and populate provided data
        """

        print("[{}] Reading ARG-GUI user settings...".format(app))

        # Clean settings
        self.settings.clear()

        # Read ADMIN level settings file
        print("[{}] Reading ARG-GUI ADMIN-lvl config file from: {}".format(app, self.adminLvlPath))
        settings = self.readSettingsFile(self.adminLvlPath)

        # Return True if admin level config file is missing, and whole list of setting keys
        if settings == False:
            return [True, argSettings]

        # Otherwise set each read setting value -- settings contains at least one element
        for key, value in settings.items():
            self.setSetting(key, value)

        # Overwrite with USER level settings file
        print("[{}] Reading ARG-GUI USER-lvl config file from: {}".format(app, self.usrLvlPath))
        settings = self.readSettingsFile(self.usrLvlPath)
        if settings:
            # Set each read setting value
            for key, value in settings.items():
                if value:
                    self.setSetting(key, value)

        # Check consistency
        missings = self.checkAllSettings()
        return [False, missings]

    @staticmethod
    def readSettingsFile(lvl):
        """Read settings file at given level
        """

        # Initiate clean state
        argErrorString = ''

        # Check provided file existence
        if not os.path.exists(lvl) or not os.path.isfile(lvl):
            argErrorString += "[{}] No config file at \'{}\'.".format(app, lvl)
            print("{}".format(argErrorString))
            return False

        # Retrieve dictionary of parameters from file
        fileLoad = None
        with open(lvl, 'r') as f:
            fileLoad = yaml.safe_load(f)

        # Bail out early if dictionary is empty
        if not fileLoad:
            argErrorString += "[{}] Could not open file at \'{}\' or empty config file.".format(app, lvl)
            print("{}".format(argErrorString))
            return {}

        return fileLoad

    def readPythonExecutable(self, lvl):
        """Read Python executable setting at given level
        """

        fileLoad = self.readSettingsFile(lvl)
        return fileLoad.get(argPythonExePathUserSettings, None)

    def readPythonSitePackage(self, lvl):
        """Read Python site-packages setting at given level
        """

        fileLoad = self.readSettingsFile(lvl)
        return fileLoad.get(argPythonSitePackagePathUserSettings, None)

    def readArgScript(self, lvl):
        """Read ARG script setting at given level
        """

        fileLoad = self.readSettingsFile(lvl)
        return fileLoad.get(argScriptPathUserSettings, None)

    def readLatexProcessor(self, lvl):
        """Read LaTeX processor setting at given level
        """

        fileLoad = self.readSettingsFile(lvl)
        return fileLoad.get(argLatexProcessorPathUserSettings, None)

    def checkAllSettings(self):
        """Check consistency of current set of settings
        """

        # Initiate list of missing settings
        missing = []

        # Check value for each setting
        if not self.getPythonExecutable():
            missing.append(argPythonExePathUserSettings)
        if not self.getPythonSitePackage():
            missing.append(argPythonSitePackagePathUserSettings)
        if not self.getArgScript():
            missing.append(argScriptPathUserSettings)
        if not self.getLatexProcessor():
            missing.append(argLatexProcessorPathUserSettings)

        # Print warning message gathering all missing settings
        if missing:
            # print("*  WARNING: Missing settings:\n\t- {}".format(",\n\t- ".join(missing)))
            return missing

    def setSetting(self, key, value):
        """Check and set provided key, value
        """

        if key in argSettings:
            self.settings[key] = value
        else:
            print("Key '{}' is not valid -- contains '{}'.".format(key, value))

    def getPythonExecutable(self):
        return self.settings[argPythonExePathUserSettings] if argPythonExePathUserSettings in self.settings else None

    def getPythonSitePackage(self):
        return self.settings[
            argPythonSitePackagePathUserSettings] if argPythonSitePackagePathUserSettings in self.settings else None

    def getArgScript(self):
        return self.settings[argScriptPathUserSettings] if argScriptPathUserSettings in self.settings else None

    def getLatexProcessor(self):
        return self.settings[
            argLatexProcessorPathUserSettings] if argLatexProcessorPathUserSettings in self.settings else None



# HEADER
#                           arg/GUI/Logic/argRunner.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import os
import re

from PySide2.QtCore import QCoreApplication, QFileInfo, QDir, QObject, QProcess, Qt, Signal

from arg.GUI.Logic.argSettingsController import argSettingsController

app = "ARG-GUI"


class argRunner(QObject):
    """A runner class
    """

    # Signals
    logStandardDetected = Signal(str)
    logErrorDetected = Signal(str)
    argRunStarted = Signal()
    argRunFinished = Signal()

    def __init__(self):

        super().__init__()

        self.argPythonExe = ""
        self.argPythonSitePackage = ""
        self.argScript = ""
        self.argLatexProcessorPath = ""
        self.argExecutionOption = None

        self.parameterFile = ""
        self.settingsController = None

        # Process creation
        self.process = QProcess()

        # Define connections
        self.process.readyReadStandardError.connect(self.onErrorDetected, Qt.DirectConnection)
        self.process.readyReadStandardOutput.connect(self.onOutputDetected, Qt.DirectConnection)

    def setSettingsController(self, settingsController: argSettingsController):
        self.settingsController = settingsController
        # Set ARG execution default option
        self.argExecutionOption = settingsController.eOption()

    def initializeEnv(self):
        """Initialize environment to be applied on ARG process started by Runner
        """

        # Initialize QProcess environment
        env = QProcess.systemEnvironment()

        # Set PYTHONPATH
        env = self.initializeEnvVar(env, "PYTHONPATH", [self.argPythonExe, self.argPythonSitePackage])

        # Set PATH
        env = self.initializeEnvVar(env, "PATH",
                                    [os.path.dirname(self.argPythonExe), self.argPythonExe,
                                     os.path.dirname(self.argLatexProcessorPath), self.argLatexProcessorPath,
                                     os.environ["PATH"] if "PATH" in os.environ else ""])

        # Update environment of process to be used to run ARG
        self.process.setEnvironment(env)

    @staticmethod
    def initializeEnvVar(env, envVarName, pathsList):
        """Initialize environment variable with provided list of paths
        """

        # Define regex
        regex = re.compile(r"^{}=(.*)".format(envVarName), re.IGNORECASE)

        # Build
        pathsStr = os.pathsep.join(pathsList)

        # TBD
        regexSub = r"{}={}".format(envVarName, os.pathsep.join(pathsList))
        env = [regex.sub(regexSub.replace('\\', '/'), var.replace('\\', '/')) for var in env]

        # Loop over system environment to look for possibly existing value
        isPathFound = False
        for var in env:
            if envVarName in var:
                isPathFound = True
        if not isPathFound:
            env.append("{}={}".format(envVarName, pathsStr))

        return env

    def updateEnv(self):
        """Initialize Runner attributes, including parsed setting values
        """

        settings = self.settingsController

        # Initialize setting values
        self.argPythonExe = settings.getArgPythonExePath()
        self.argPythonSitePackage = settings.getArgPythonSitePackagePath()
        self.argScript = settings.getArgScriptPath()
        self.argLatexProcessorPath = settings.getArgLatexProcessorPath()

        # Initialize ARG process environement
        self.initializeEnv()

    def start(self):

        # Update ARG process environement
        self.updateEnv()

        # Retrieve settings controller
        settings = self.settingsController

        # Retrieve current file
        self.parameterFile = settings.getCurrentParameterFileRun()

        # Initiate ARG process if current file defined
        if self.parameterFile:

            # Set working directory
            parameterFileDir = QFileInfo(self.parameterFile).canonicalPath()
            self.process.setWorkingDirectory(parameterFileDir)
            print("[{}] Working directory: {}".format(app, parameterFileDir))

            # Gather arguments to run ARG with
            arguments = [settings.getArgScriptPath(), self.argExecutionOption, "-p", self.parameterFile]
            if self.argLatexProcessorPath:
                arguments.append("-l")
                arguments.append(self.argLatexProcessorPath)
            print("[{}] Starting with {} {}".format(app, settings.getArgPythonExePath(), arguments))

            # Emit start of process signal
            self.argRunStarted.emit()

            # Run ARG process
            self.process.start(settings.getArgPythonExePath(), arguments)
            self.process.waitForFinished(-1)
            # print("[{}] Ending: {}".format(app, self.process.readAll()))
            # print("[{}] Error: {}".format(app, self.process.readAllStandardError()))

            # Emit end of process signal
            self.argRunFinished.emit()

        # Ask for parameters file otherwise
        else:
            self.logErrorDetected.emit("Please, set a parameters file to run.")

        return 1

    def clean(self, outputFolder):
        settings = QCoreApplication.instance().settingsController

        # Get current file
        parameterFile = settings.getCurrentParameterFile()
        if parameterFile:
            # Set working directory
            parameterFileDir = QFileInfo(parameterFile).canonicalPath()

            if outputFolder:
                if os.path.exists(outputFolder) and os.path.isdir(outputFolder):
                    outputFolderFullPath = outputFolder
                else:
                    outputFolderFullPath = os.path.join(parameterFileDir, outputFolder)
                outputFolderFullPathAsDir = QDir(outputFolderFullPath)
                if outputFolderFullPathAsDir.removeRecursively():
                    self.logStandardDetected.emit(
                        "[{}] 'Clean' action - the following output folder has been cleaned: {}"
                            .format(app, outputFolderFullPath))
                else:
                    self.logErrorDetected.emit(
                        "** ERROR: 'Clean' action - the following output folder cannot be cleaned: {}"
                            .format(outputFolderFullPath))
            else:
                self.logErrorDetected.emit(
                    "** ERROR: 'Clean' action - the following output folder cannot be cleaned: {}"
                        .format(outputFolder))

    def setEOption(self):
        """ Switch runner to E execution mode
        """
        self.argExecutionOption = self.settingsController.eOption()

    def setGOption(self):
        """ Switch runner to G execution mode
        """
        self.argExecutionOption = self.settingsController.gOption()

    def onErrorDetected(self):
        self.logErrorDetected.emit(str(self.process.readAllStandardError(), 'utf-8'))

    def onOutputDetected(self):
        self.logStandardDetected.emit(str(self.process.readAllStandardOutput(), 'utf-8'))



#HEADER
#                    arg/Applications/argExplorator.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

from arg.Tools import Utilities
from arg.Generation import argPlot, argVTK

class argExploratorSolution:
    """A class to encapsulate solution parameters
    """

    def __init__(self):
        # Default instance variables
        self.Type = None
        self.Name = None
        self.Method = None

class argExplorator:
    """A class to explore data and create structure file
    """

    def __init__(self, parameters):
        """ Constructor
        """

        # Data directory and files provided by user
        self.DataDir = parameters.DataDir
        self.DataPath = os.path.join(parameters.DataDir, '')
        self.ParametersFile = parameters.ParametersFile
        self.Mutables = parameters.Mutables
        self.RealDataDir = os.path.realpath(parameters.DataDir)

        # Other parameters inherited from user specifications
        self.Mappings = parameters.Mappings
        self.MetaData = parameters.MetaData
        self.IgnoredBlockKeys = parameters.IgnoredBlockKeys
        self.SolutionCases = parameters.SolutionCases
        self.Fragments = parameters.Fragments

        # Discovered files
        self.DeckFiles = []
        self.ExodusIIFiles = []
        self.ExodusIIPartitions = set()
        self.LogNames = []
        self.Images = []
        self.TextFiles = []

        # Variables determined by analysis
        self.DiscoveredData = {}
        self.GeometryFiles = []
        self.ParametersFiles = []
        self.DeckType = None
        self.DeckRoot = None
        self.MeshType = None
        self.MeshName = None
        self.LogFile = None
        self.Solutions = []
        self.CaseNames = []
        self.Methods = []

    def print(self):
        """ Print debug information
        """

        # Initialize empty print content
        res = ''

        # Data directory and files provided by user
        res = "{}\nDataDir = {}".format(res, self.DataDir)
        res = "{}\nDataPath = {}".format(res, self.DataPath)
        res = "{}\nParametersFile = {}".format(res, self.ParametersFile)
        res = "{}\nMutables = {}".format(res, self.Mutables)
        res = "{}\nRealDataDir = {}".format(res, self.RealDataDir)

        # Other parameters inherited from user specifications
        res = "{}\nMappings = {}".format(res, self.Mappings)
        res = "{}\nMetaData = {}".format(res, self.MetaData)
        res = "{}\nIgnoredBlockKeys = {}".format(res, self.IgnoredBlockKeys)
        res = "{}\nSolutionCases = {}".format(res, self.SolutionCases)
        res = "{}\nFragments = {}".format(res, self.Fragments)

        # Discovered files
        res = "{}\nDeckFiles = {}".format(res, self.DeckFiles)
        res = "{}\nExodusIIFiles = {}".format(res, self.ExodusIIFiles)
        res = "{}\nExodusIIPartitions = {}".format(res, self.ExodusIIPartitions)
        res = "{}\nLogNames = {}".format(res, self.LogNames)
        res = "{}\nImages = {}".format(res, self.Images)
        res = "{}\nTextFiles = {}".format(res, self.TextFiles)

        # Variables determined by analysis
        res = "{}\nDiscoveredData = {}".format(res, self.DiscoveredData)
        res = "{}\nGeometryFiles = {}".format(res, self.GeometryFiles)
        res = "{}\nParametersFiles = {}".format(res, self.ParametersFiles)
        res = "{}\nDeckType = {}".format(res, self.DeckType)
        res = "{}\nDeckRoot = {}".format(res, self.DeckRoot)
        res = "{}\nMeshType = {}".format(res, self.MeshType)
        res = "{}\nMeshName = {}".format(res, self.MeshName)
        res = "{}\nLogFile = {}".format(res, self.LogFile)
        res = "{}\nSolutions = {}".format(res, self.Solutions)
        res = "{}\nCaseNames = {}".format(res, self.CaseNames)
        res = "{}\nMethods = {}".format(res, self.Methods)

        # Print content
        print(res)

    def recursively_search_supported_files(self, current_dir, verbosity):
        """ Recursively retrieve supported files in given directory tree
        """

        # Iterate over all files in all current directory
        for file_name in os.listdir(current_dir):
            dir_or_file = os.path.join(current_dir, file_name)

            # Distinguish between sub-directories and regular files
            if os.path.isdir(dir_or_file):
                # Descend into sub-tree
                self.recursively_search_supported_files(dir_or_file, verbosity)

            elif os.path.isfile(dir_or_file):
                # Split file name
                file_split = file_name.split('.')
                if len(file_split) > 1:
                    last_extension = file_split[-1].lower()
                    if len(file_split) > 3:
                        ante_penult_extension = file_split[-3].lower()
                    else:
                        ante_penult_extension = ''
                else:
                    last_extension = ''
                    ante_penult_extension = ''

                # Handle all supported regular file types
                file_name = dir_or_file.replace(self.DataPath, '', 1)
                if last_extension in ('i', "in", "inp", "yml", "yaml"):
                    # Potential input deck file
                    if file_name not in (self.ParametersFile, self.Mutables):
                        self.DeckFiles.append(file_name)
                elif last_extension in ('e', 'g', "ex2", "exo"):
                    # ExodusII files
                    self.ExodusIIFiles.append(file_name)
                elif ante_penult_extension in ('e', 'g', "ex2", "exo"):
                    # ExodusII partitions
                    self.ExodusIIPartitions.add(os.path.splitext(file_name)[0])
                elif last_extension in ("log", "rslt"):
                    # Log files
                    self.LogNames.append(file_name)
                elif verbosity:
                    if last_extension == "png":
                        # Stand-alone images
                        print("[Explorator] Found stand-alone image".format(
                            file_name))
                        self.Images.append(file_name)
                    elif last_extension == "txt":
                        # Stand-alone text fragments
                        print("[Explorator] Found stand-alone text fragment".format(
                            file_name))
                        self.TextFiles.append(file_name)



#HEADER
#                      arg/GUI/Logic/argParameterReader.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import yaml
from PySide2.QtCore import QObject, Signal

from arg.GUI.Logic.argParameterData import argParameterData
from arg.GUI.Logic.argSettingsController import argSettingsController

app = "ARG-GUI"

Types = {"BackendTypes": {
    "LaTeX": {
        "Main": "tex",
        "Captions": "tex"},
    "Word": {
        "Main": "txt",
        "Captions": "txt"}},
    "ReportTypes": {
        "Report": "Generic"},
    "ClassificationLevels": {
        "Generic": {}},
    "VerbosityLevels": {
        "terse": -1,
        "default": 0,
        "verbose": 1}}


class argParameterReader(QObject):
    """A reader class to read parameters file
    """

    # Signals
    dataCreated = Signal(dict)
    errorDetected = Signal(str)

    def __init__(self, data=None, err=""):
        super().__init__()

        if data is None:
            data = dict()

        self.ParameterData = argParameterData(data)
        self.ErrorString = err
        self.SettingController = argSettingsController()

    @staticmethod
    def verbosity_to_int(value):
        """Convert verbosity key into integer value in allowable range
        """

        # Retrieve verbosity levels
        verbosity_levels = Types.get("VerbosityLevels", {})

        # Try to convert passed value to integer
        try:
            # If an integer was passed, verify that it is in range
            int_value = int(value)
            if int_value in verbosity_levels.values():
                # Integer value is in range, nothing to do
                return int_value
            else:
                # Integer value is in not in range, warn and return default
                default_verb = verbosity_levels.get("default")
                print(
                    "*  WARNING: {} is not a valid verbosity integer identifier, assigning default ({}) instead".format(
                        value,
                        default_verb))
                return default_verb
        except:
            # Otherswise check that is one of the allowed keys
            int_value = verbosity_levels.get(value)
            if int_value is None:
                # String key is undefined, warn and return default
                key = "default"
                default_verb = verbosity_levels.get(key)
                print("*  WARNING: {} is not a valid verbosity string identifier, assigning {} instead".format(
                    value,
                    key))
                return default_verb
            else:
                # String key is defined, returned corresponding integer
                return int_value

    def read(self, filePath):
        """Read specified file and populate provided data
        """

        # Initiate clean state
        dictTmp = {}
        argErrorString = ""

        # Retrieve dictionary of parameters from file
        fileLoad = None
        with open(filePath) as f:
            fileLoad = yaml.safe_load(f)

        # Bail out early if dictionary is empty
        if not fileLoad:
            argErrorString += "[{}] No parameters file at {}".format(app, filePath)
            self.errorDetected.emit(argErrorString)
            return False

        # Assemble possible report classification keys
        class_keys = {}
        for level, level_keys in Types.get("ClassificationLevels", {}).items():
            for key, values in level_keys.items():
                compound_key = "{}_{}".format(level, key)
                class_keys[compound_key.lower()] = [compound_key, values]

        # Fill default value
        dictTmp[self.SettingController.finalKey()] = False
        dictTmp[self.SettingController.verbosityKey()] = self.SettingController.verbosityKeyDefaultAsInt()

        # Process retrieved non-empty dictionary
        for key, value in fileLoad.items():

            # Keys are case-insensitive
            key = key.lower()

            # Start with Boolean statements
            if key == self.SettingController.finalParameterFileKey() and value:
                dictTmp[self.SettingController.finalKey()] = True

            # Then handle all straightforward key: value statements
            elif key == self.SettingController.fileNameParameterFileKey():
                dictTmp[self.SettingController.fileNameKey()] = value
            elif key == self.SettingController.mutablesFileParameterFileKey():
                dictTmp[self.SettingController.mutablesFileKey()] = value
            elif key == self.SettingController.dataFolderParameterFileKey():
                dictTmp[self.SettingController.dataFolderKey()] = value
            elif key == self.SettingController.outputFolderParameterFileKey():
                dictTmp[self.SettingController.outputFolderKey()] = value
            elif key == self.SettingController.structureFileParameterFileKey():
                dictTmp[self.SettingController.structureFileKey()] = value
            elif key == self.SettingController.structureEndFileParameterFileKey():
                dictTmp[self.SettingController.structureEndFileKey()] = value
            elif key == self.SettingController.yearParameterFileKey():
                dictTmp[self.SettingController.yearKey()] = value
            elif key == self.SettingController.monthParameterFileKey():
                dictTmp[self.SettingController.monthKey()] = value
            elif key == self.SettingController.numberParameterFileKey():
                dictTmp[self.SettingController.numberKey()] = value
            elif key == self.SettingController.issueParameterFileKey():
                dictTmp[self.SettingController.issueKey()] = value
            elif key == self.SettingController.versionParameterFileKey():
                if self.SettingController.versionsKey() in dictTmp:
                    dictTmp[self.SettingController.versionsKey()].append(value.split(','))
                else:
                    dictTmp[self.SettingController.versionsKey()] = value
            elif key == self.SettingController.titleParameterFileKey():
                dictTmp[self.SettingController.titleKey()] = value
            elif key == self.SettingController.authorParameterFileKey():
                if self.SettingController.authorsKey() in dictTmp:
                    dictTmp[self.SettingController.authorsKey()].append(value)
                else:
                    dictTmp[self.SettingController.authorsKey()] = value
            elif key == self.SettingController.organizationParameterFileKey():
                if self.SettingController.organizationsKey() in dictTmp:
                    dictTmp[self.SettingController.organizationsKey()].append(value)
                else:
                    dictTmp[self.SettingController.organizationsKey()] = value
            elif key == self.SettingController.abstractFileParameterFileKey():
                dictTmp[self.SettingController.abstractFileKey()] = value
            elif key == self.SettingController.thanksParameterFileKey():
                dictTmp[self.SettingController.thanksKey()] = value
            elif key == self.SettingController.prefaceParameterFileKey():
                dictTmp[self.SettingController.prefaceKey()] = value
            elif key == self.SettingController.executiveSummaryParameterFileKey():
                dictTmp[self.SettingController.executiveSummaryKey()] = value
            elif key == self.SettingController.nomenclatureParameterFileKey():
                dictTmp[self.SettingController.nomenclatureKey()] = value
            elif key == self.SettingController.locationParameterFileKey():
                dictTmp[self.SettingController.locationKey()] = value
            elif key == self.SettingController.artifactFileParameterFileKey():
                dictTmp[self.SettingController.artifactFileKey()] = value
            elif key == self.SettingController.geometryRootParameterFileKey():
                dictTmp[self.SettingController.geometryRootKey()] = value
            elif key == self.SettingController.inputDeckParameterFileKey():
                dictTmp[self.SettingController.inputDeckKey()] = value
            elif key == self.SettingController.logFileParameterFileKey():
                dictTmp[self.SettingController.logFileKey()] = value

                # Then interpret verbosity key as integer value
            elif key == self.SettingController.verbosityParameterFileKey():
                dictTmp[self.SettingController.verbosityKey()] = self.verbosity_to_int(value)

            # Key separator value must be a single character
            elif key == self.SettingController.keySeparatorParameterFileKey() and isinstance(value, str) and len(
                    value) == 1:
                dictTmp[self.SettingController.keySeparatorKey()] = value

            # Report type
            if key == self.SettingController.reportTypeParameterFileKey():
                allow = Types.get("ReportTypes", [])
                if value in allow:
                    dictTmp[self.SettingController.reportTypeKey()] = value
                else:
                    argErrorString += "** ERROR: `{}` is not a valid report type. Allowed values are in {}".format(
                        value,
                        allow)

            # Backend type
            if key == self.SettingController.backendTypeParameterFileKey():
                allow = Types.get("BackendTypes", {})
                if value in allow:
                    dictTmp[self.SettingController.backendTypeKey()] = value
                else:
                    argErrorString += "** ERROR: `{}` is not a valid backend type. Allowed values are in {}".format(
                        value,
                        allow)

            # Classification
            elif key == self.SettingController.classificationParameterFileKey():
                class_type = [x for x in Types.get("ClassificationLevels", {}).keys() if x.lower() == value.lower()]
                if not class_type:
                    argErrorString += "** ERROR: `{}` is not a valid classification. Allowed values are in {}".format(
                        value,
                        Types.get("ClassificationLevels", {}).keys())
                else:
                    next_type = next((x for x in class_type), None)
                    if next_type:
                        dictTmp[self.SettingController.classificationKey()] = next_type
                    else:
                        dictTmp[self.SettingController.classificationKey()] = None

            # Classification sub-key
            elif key in class_keys:

                restricted_values = class_keys[key][1]
                if not restricted_values or (
                        restricted_values and value in restricted_values):
                    setattr(self, class_keys[key][0], value)

                # Report on invalid values and error out
                elif restricted_values:
                    argErrorString += "** ERROR: `{}` is not allowed for `{}`. Allowed values are in {}".format(
                        value,
                        class_keys[key][0],
                        restricted_values)

            # Handle key=list of values statements
            elif key in (
                    self.SettingController.reportedCadMetaDataParameterFileKey(),
                    self.SettingController.ignoredBlocksParameterFileKey(),
                    self.SettingController.solutionCasesParameterFileKey(),
                    self.SettingController.insertionParameterFileKey(),
            ):
                if not isinstance(value, list):
                    print("*  WARNING: ill-formed {} directive: a {} was passed instead of a list".format(
                        key,
                        type(value)))
                    continue
                if key == self.SettingController.reportedCadMetaDataParameterFileKey():
                    dictTmp[self.SettingController.reportedCadMetaDataKey()] = ["{}".format(x) for x in value]
                elif key == self.SettingController.solutionCasesParameterFileKey():
                    dictTmp[self.SettingController.solutionCasesKey()] = value
                elif key == self.SettingController.ignoredBlocksParameterFileKey():
                    dictTmp[self.SettingController.ignoredBlocksKey()] = ["{}".format(x).lower() for x in value]
                elif key == self.SettingController.insertionParameterFileKey():
                    dictTmp[self.SettingController.insertionKey()] = value

            # Handle key=dictionary of values statements
            elif key in (
                    self.SettingController.bijectiveMappingsParameterFileKey(),
            ):
                if not isinstance(value, dict):
                    print("*  WARNING: ill-formed {} directive: a {} was passed instead of a dict".format(
                        key,
                        type(value)))
                    continue
                elif key == self.SettingController.bijectiveMappingsParameterFileKey():
                    dictTmp[self.SettingController.bijectiveMappingsKey()] = value

        # Return True and emit dict if no error
        if argErrorString == "":
            self.ParameterData = argParameterData(dictTmp)
            self.dataCreated.emit(self.ParameterData.Dict)
            self.ErrorString = argErrorString
            return True

        # Otherwise, emit error string
        else:
            self.ErrorString = argErrorString
            self.errorDetected.emit(self.ErrorString)
            # TO RM After properly send it to the application
            print("[{}] Could not properly read specified parameters file content: {}".format(app, argErrorString))
            return False



#HEADER
#                          arg/GUI/View/__init__.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER


#HEADER
#                  arg/Common/argInformationObject.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

from collections.abc import Hashable


class argInformationObject:
    """ A class to store and exchange information between ARG components
    """

    def __init__(self, info_type):
        """ Constructor
        """

        # Keep track of provided information type
        self.InformationType = info_type

        # Initialize list of information names
        self.InformationNames = []

        # Initialize internal map
        self.InformationMap = {}

    def get_type(self):
        """ Return information type
        """

        return self.InformationType

    def get_names(self):
        """ Return information names
        """

        return self.InformationNames

    def set_names(self, info_names):
        """ Set information names
        """

        if isinstance(info_names, list):
            self.InformationNames = info_names
        else:
            print("*  WARNING: attempted to set non-list {} type as argInformationObject information names. Ignoring it.".format(
                type(info_names)))

    def clear(self):
        """ Clear internal storage
        """

        # Clear internal storage
        self.InformationMap.clear()

    def pop(self, value=None):
        """ Pop last value from internal storage
        """

        # Safely pop value from internal storage
        return self.InformationMap.pop(value, None)

    def update(self, info_key, info_value):
        """ Try to update internal map with key and value pair
        """

        # Bail out if a non-hashable type was passed as first input
        if not isinstance(info_key, Hashable):
            print("*  WARNING: attempted to append a non-hashable {} type to argInformationObject. Ignoring it.".format(
                type(info_key)))
            return

        # Update internal dictionary
        self.InformationMap.update({info_key: info_value})

    def dump(self):
        """ Print contents of internal storage for debugging purposes
        """

        # Print internal content
        print("{} contains: {}".format(self, (", ".join([str((k, v)) for k, v in self.InformationMap.items()]))))

    def iterator(self):
        """ Provide iterator over internals
        """

        # Create generator over list of doublets
        for info_key, info_value in self.InformationMap.items():
            yield info_key, info_value



# HEADER
#                        arg/GUI/Logic/argApplication.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import datetime
import os
import sys

from PySide2.QtCore import QFile, QFileInfo, QSettings, Slot, Qt
from PySide2.QtWidgets import QApplication, QMessageBox

from arg.GUI.Logic.argActionManager import argActionManager
from arg.GUI.Logic.argParameterController import argParameterController
from arg.GUI.Logic.argRunner import argRunner
from arg.GUI.Logic.argSettingsController import argSettingsController
from arg.GUI.View.argUserSettingsDialog import argUserSettingsDialog


class argApplication(QApplication):
    """An application class
    """

    def __init__(self, parent=None):
        super().__init__()

        # Create controllers, runner and action manager
        self.settingsController = argSettingsController()

        # Retrieve list of missing settings from settings controller initialization
        missingSettings = self.settingsController.initialize()

        # Show missing settings if there are any
        if missingSettings:
            self.showConfigMissingSettings(missingSettings, True)
        self.parameterController = argParameterController()
        self.parameterController.setSettingsController(self.settingsController)
        self.runner = argRunner()
        self.runner.setSettingsController(self.settingsController)
        self.actionManager = argActionManager()

        # Pointer on mainWindow
        self.mainWindow = None

        # User setting dialog
        self.userSettingsDialog = argUserSettingsDialog()
        self.userSettingsDialog.setModal(True)

        # Save before run
        permanentSettings = QSettings(QSettings.IniFormat, QSettings.UserScope,
                                      self.settingsController.getCompanyName(),
                                      self.settingsController.getToolName())
        self.saveBeforeRunEnabled = permanentSettings.value(self.settingsController.getSaveBeforeRunSettings(), False,
                                                            bool)
        self.actionManager.saveBeforeRunAct.setChecked(self.saveBeforeRunEnabled)

        # Connection
        self.parameterController.dataCreated.connect(self.onDataCreated)
        self.runner.logErrorDetected.connect(self.onLogErrorDetected)
        self.runner.logStandardDetected.connect(self.onLogStandardDetected)
        self.runner.argRunStarted.connect(self.onArgRunStarted)
        self.runner.argRunFinished.connect(self.onArgRunFinished)

        # Dialog Message box
        self.messageBox = QMessageBox()
        self.messageBox.setModal(True)

        self.setApplicationName("ARG-GUI")

    def setApplicationTitle(self, parameterFile):
        """Set the application name using title parameter
        """
        title = self.applicationName()
        if parameterFile != "":
            title = "{} - {}".format(title, parameterFile)
        else:
            title = "{} - {}".format(title, "Untitled")

        self.mainWindow.setWindowTitle(title)

    def setMainWindow(self, window):
        """Main window setter
        """

        self.mainWindow = window

    def openRequested(self, filePath):
        """Upon open action request
        """

        # Read file at provided path
        if self.parameterController.read(filePath):
            self.mainWindow.addRecentFileOpenAction(filePath)
            # Change path in title bar if properly read
            self.setApplicationTitle(self.settingsController.getCurrentParameterFile())
        # Log an error otherwise
        else:
            print("** ERROR: Failed to open parameters file: '{}'".format(filePath))

    @staticmethod
    def quitRequested():
        """Upon quit action request
        """
        sys.exit()

    def saveRequested(self, filePath):
        """Upon save action request
        """

        # Construct data on the basis of current displayed values
        data = self.mainWindow.constructParameters()

        # Call 'Save' action on provided file path
        if self.parameterController.write(filePath, data):
            self.setApplicationTitle(self.settingsController.getCurrentParameterFile())
            self.messageBox.setText("{} has been saved".format(filePath))
            self.messageBox.show()

    def saveBeforeRunRequested(self, enabled):
        self.saveBeforeRunEnabled = enabled

    def runRequested(self):
        """Upon run action request
        """

        # Clean logger content
        self.mainWindow.clearLogs()

        if self.saveBeforeRunEnabled:
            self.actionManager.saveAct.trigger()

        # Retrieve data from parameters values
        data = self.mainWindow.constructParameters()

        # Get current file
        parameterFile = self.settingsController.getCurrentParameterFile()

        # Call 'Run' action if current file defined
        if parameterFile:

            # Save current parameters values in temp file
            parameterQFile = QFileInfo(parameterFile)
            fileCanonicalFilePath = parameterQFile.canonicalPath()
            fileBaseName = parameterQFile.baseName()
            fileSuffix = parameterQFile.suffix()
            fileDateTime = datetime.datetime.now().strftime("-%Y-%m-%d-%H-%M-%S")
            filePathTemp = "{}/{}{}.{}".format(fileCanonicalFilePath, fileBaseName, fileDateTime, fileSuffix)
            self.parameterController.write(filePathTemp, data)
            self.settingsController.setCurrentParameterFileRun(filePathTemp)

            # Write overwrite the currentParameter file, but here, it is done with a temp file
            # As a consequence, we really need to give it is previous and correct value
            self.settingsController.setCurrentParameterFile(parameterFile)
            self.runner.start()

        # Ask for parameters file otherwise
        else:
            self.onLogErrorDetected("Cannot run without a parameters file. ")

    def cleanRequested(self):
        """Upon clean action request
        """
        # Retrieve data from parameters values
        data = self.mainWindow.constructParameters()
        outputFolder = ""

        if self.settingsController.outputFolderKey() in data:
            outputFolder = data.get(self.settingsController.outputFolderKey())

        # Call the clean method of the runner
        self.runner.clean(outputFolder)

    def reloadRequested(self):
        """Upon reload action request
        """

        # Reload backup data, as read from current parameters file initial state
        self.parameterController.reloadData()

    @staticmethod
    def helpRequested():
        """Upon help action request
        """

        # Create new message box
        helpWindow = QMessageBox()
        helpWindow.setWindowTitle("Help on parameters file")
        helpWindow.setModal(True)

        # Flags: disable minimize and close buttons
        helpWindow.setWindowFlags(Qt.Tool | Qt.WindowTitleHint | Qt.CustomizeWindowHint)

        # Path to Word documentation -- contained in repository
        helpWindow.setText("Word documentation on parameters file is located in '<code>arg\\doc\\user_manual</code>'."
                           " <br><br> Please refer to '<code>[User Manual] How to create a parameters.yml "
                           "file_0.4.0.docx</code>'. ")
        helpWindow.setTextFormat(Qt.RichText)
        helpWindow.addButton(QMessageBox.Ok)

        helpWindow.exec()

    @staticmethod
    def aboutRequested():
        """Upon help action request
        """

        # Create new message box
        aboutWindow = QMessageBox()
        aboutWindow.setWindowTitle("About ARG-GUI")
        aboutWindow.setModal(True)

        # Flags: disable minimize and close buttons
        aboutWindow.setWindowFlags(Qt.Tool | Qt.WindowTitleHint | Qt.CustomizeWindowHint)

        # Path to Word documentation -- contained in repository
        aboutWindow.setText("ARG is an Automatic Report Generator. ")
        aboutWindow.setTextFormat(Qt.RichText)
        aboutWindow.addButton(QMessageBox.Ok)

        aboutWindow.exec()

    def openUserSettings(self):
        """Upon open user settings request
        """

        self.userSettingsDialog.initialize()
        self.userSettingsDialog.show()

    def runEButtonClicked(self):
        """ Switch runner to E execution mode
        """

        qapp = QApplication.instance()
        self.runner.setEOption()

    def runGButtonClicked(self):
        """ Switch runner to G execution mode
        """
        qapp = QApplication.instance()
        self.runner.setGOption()

    @Slot(dict)
    def onDataCreated(self, data):
        """Dictionary slot receiving created data
        """

        self.mainWindow.fillParameters(data)

    @Slot(str)
    def onLogStandardDetected(self, log):
        """String slot receiving standard log message
        """

        self.mainWindow.logStandard(log)

    @Slot(str)
    def onLogErrorDetected(self, log):
        """String slot receiving error log message
        """

        self.mainWindow.logError(log)

    @Slot()
    def onArgRunStarted(self):
        """Slot receiving ARG run state: currently not implemented
        """

        pass

    @Slot()
    def onArgRunFinished(self):
        """Slot receiving ARG run state
        """

        parameterFile = self.settingsController.getCurrentParameterFileRun()
        currentRunFile = QFile(parameterFile)
        currentRunFile.remove()
        self.settingsController.setCurrentParameterFileRun("")

    @staticmethod
    def showConfigMissingSettings(missings, exitEarly):
        """Create a message box listing all missing settings
        """
        app = 'ARG-GUI'
        # Print error message for log
        print("[{}] Following settings are missing:\n" \
              "\t- \'{}\'\nPlease define at USER or ADMIN level. Exiting.".format(app, "\'\n\t- \'".join(missings)))

        # Create new message box
        missingsWindow = QMessageBox()
        missingsWindow.setWindowTitle("ARG-GUI missing settings")
        missingsWindow.setModal(True)

        # Flags: disable minimize and close buttons
        missingsWindow.setWindowFlags(Qt.Tool | Qt.WindowTitleHint | Qt.CustomizeWindowHint)

        # Path to Word documentation -- contained in repository
        missingsWindow.setText("Following settings are missing:<br>" \
                               "- \'<code>{}</code>\'<br>Please define at USER or ADMIN level. Exiting.".format(
            "</code>\'<br>- \'<code>".join(missings)))
        missingsWindow.setTextFormat(Qt.RichText)
        missingsWindow.setIcon(QMessageBox.Critical)
        missingsWindow.addButton(QMessageBox.Ok)

        # Show created windows
        missingsWindow.exec()

        # Exit early
        if exitEarly:
            sys.exit(1)

    def checkLineEdit(self, lineEdit, isDirectory):
        # Check if the content of the linEdit exists.
        # if isDirectory, the content is considered as a directory.
        # if not, the content is considered as a text
        # An empty string is considered as validated
        self.checkLineEditRecursive(lineEdit, isDirectory, lineEdit.text())

    def checkLineEditRecursive(self, lineEdit, isDirectory, filePath):
        # Check if the content of the linEdit exists.
        # if isDirectory, the content is considered as a directory.
        # if not, the content is considered as a text
        # An empty string is considered as validated
        fileInfo = QFileInfo(filePath)
        if lineEdit.text() == "":
            # Good case: Empty string
            lineEdit.setStyleSheet(self.getStyleSheetLineEditNormal())
        else:
            if fileInfo.exists():
                isTypeOk = False
                if (isDirectory and fileInfo.isDir()) or (not isDirectory and fileInfo.isFile()):
                    isTypeOk = True
                if isTypeOk:
                    # Good case: Type ok
                    lineEdit.setStyleSheet(self.getStyleSheetLineEditNormal())
                else:
                    # Wrong case: Type no ok
                    lineEdit.setStyleSheet(self.getStyleSheetLineEditError())
            else:
                # Wrong case: FileInfo doesn't exist
                currentParamFilePath = QFileInfo(self.settingsController.getCurrentParameterFile()).canonicalPath()
                if not currentParamFilePath in filePath:
                    self.checkLineEditRecursive(lineEdit, isDirectory, os.path.join(currentParamFilePath, filePath))
                else:
                    lineEdit.setStyleSheet(self.getStyleSheetLineEditError())

    @staticmethod
    def getStyleSheetLineEditError():
        return "QLineEdit { background: rgb(255, 0, 0);}"

    @staticmethod
    def getStyleSheetLineEditNormal():
        return "QLineEdit {}"



#HEADER
#                       arg/GUI/View/argCentralWidget.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

from PySide2.QtWidgets import QSplitter

from arg.GUI.View.argLoggerWidget import argLoggerWidget
from arg.GUI.Logic.argParametersWidget import argParametersWidget


class argCentralWidget(QSplitter):
    """A widget class to cover central area layout
    """

    def __init__(self):
        super().__init__()

        # Instantiate 'Parameters' and 'Logger' widgets
        self.parameterWidget = argParametersWidget()
        self.loggerWidget = argLoggerWidget()

        # Add those high-level widget to central layout
        self.addWidget(self.parameterWidget)
        self.addWidget(self.loggerWidget)

    def fillParameters(self, data):
        """Fill central area widgets with provided data
        """
        # self.parameterWidget.updateTabs(data)
        self.parameterWidget.fillParameters(data)

    def constructParameters(self):
        """Instantiate central area widgets and construct corresponding data
        """

        return self.parameterWidget.constructParameters()

    def logStandard(self, log):
        """Fill log widget with provided standard log message
        """

        self.loggerWidget.logStandard(log)

    def logError(self, log):
        """Fill log widget with provided error log message
        """

        self.loggerWidget.logError(log)

    def clearLogs(self):
        """Clear log from widget
        """

        self.loggerWidget.clearLogs()

    def runEButtonClicked(self):
        """ Switch to E execution mode
        """

        self.parameterWidget.runEButtonClicked()

    def runGButtonClicked(self):
        """ Switch to G execution mode
        """

        self.parameterWidget.runGButtonClicked()



#HEADER
#                         arg/GUI/Logic/argOpenAction.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QAction, QApplication, QFileDialog

app = "ARG-GUI"


class argOpenAction(QAction):
    """An action class
    """

    def __init__(self):
        super().__init__()
        scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        self.setIcon(QIcon("{}/{}".format(scriptDirectory, "../Graphics/open.png")))
        self.setText("Open...")
        self.setToolTip("Open existing parameters file - Must be in YAML format")

        self.triggered.connect(self.onTriggered)

    @staticmethod
    def onTriggered():
        print("[{}] 'Open' action detected.".format(app))

        openFileDialog = QFileDialog()
        openFileDialog.setFileMode(QFileDialog.AnyFile)
        # openFileDialog.setFilter(QDir.NoDotAndDotDot)
        openFileDialog.setViewMode(QFileDialog.Detail)
        openFileDialog.setNameFilter("yaml file (*.yml *.yaml)")
        if openFileDialog.exec_():
            fileNames = openFileDialog.selectedFiles()
            if len(fileNames) == 1:
                QApplication.instance().openRequested(fileNames[0])
            # else: TODO: log the error.



#HEADER
#                      arg/Backend/argBackendBase.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import abc
import datetime
import os
import shutil
import sys
import yaml

from arg.Common.argInformationObject import argInformationObject
from arg.Common.argMultiFontStringHelper import argMultiFontStringHelper
from arg.Aggregation import argExodusAggregator, argVTKSTLAggregator

class argBackendBase:
    """A backend abstractbase class
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self, parameters=None):

        # Backend type to be provided by concrete subclasses
        self.Type = None

        # Keep track of specified report parameters
        self.Parameters = parameters

        # Internal storage for the report
        self.Report = None

        # Load supported types
        common_dir = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(common_dir, "../Common/argTypes.yml"),
                  'r',
                  encoding="utf-8") as t_file:
            self.Types = yaml.safe_load(t_file)

        # Retrieve supported verbosity levels
        self.VerbosityLevels = self.Types.get("VerbosityLevels", {})

        # Retrieve supported templates information
        self.Templates = self.Types.get("TemplateFiles", {})

    def get_type(self):
        """Convenience method to get backend type
        """

        # Return value of instance variable
        return self.Type

    def get_caption_extension(self):
        """Convenience method to get caption file extension for backend
        """

        # Return value of instance variable
        return argMultiFontStringHelper.Types.get(
            "BackendTypes", {}).get(self.Type, {}).get("captions", '')

    @staticmethod
    def get_timestamp():
        """Convenience method to get timestamp under unified format
        """

        # Return value of instance variable
        return datetime.datetime.strftime(datetime.datetime.now(), "%Y-%m-%d, %H:%M:%S")

    def create_document_preamble(self, version=None):
        """Must be overwritten if concrete backend needs a preamble
        """

        # No default preamble is provided
        return

    def add_document_provenance(self, version=None):
        """Must be overwritten if concrete backend needs a document timestamp
        Add document provenance information on a new page
        """

        # No default preamble is provided
        return

    def add_document_tocs(self, version=None):
        """Must be overwritten if concrete backend needs Table of Contents, List of Figures and List of Tables
        Add document Table of Contents, List of Figures and List of Tables on new pages
        """

        # No default preamble is provided
        return

    def append_document_postamble(self):
        """Must be overwritten if concrete backend needs a postamble
        """

        # No default postamble is provided
        return

    @abc.abstractmethod
    def generate_multi_font_string(self, multi_font_string, handle=None):
        """Generate backend-end specific artifact for multi-font string
        """

    @abc.abstractmethod
    def generate_matrix_string(self, matrix):
        """Generate back-end specific string from matrix entries
        """

    @abc.abstractmethod
    def add_comment(self,  comments_dict, key):
        """Add comment to the report from a dict as either text string
        or as sub-paragraph depending on number of dict values (1 or 2)
        """

    @abc.abstractmethod
    def add_paragraph(self, item):
        """Add paragraph to the report
        """

    @abc.abstractmethod
    def add_list(self, item, number_items=False):
        """Add itemization or enumeration to the report
        """

    @abc.abstractmethod
    def add_subsection(self, item, numbered=True):
        """Add subsection to the report
        """

    @abc.abstractmethod
    def add_section(self, item, numbered=True):
        """Add section to the report
        """

    @abc.abstractmethod
    def add_chapter(self, item, numbered=True):
        """Add chapter to the report
        """

    @abc.abstractmethod
    def add_page_break(self):
        """Add page break to the report
        """

    @abc.abstractmethod
    def add_table(self, header, body, do_verbatim, position):
        """Add table to the report
        """

    @abc.abstractmethod
    def add_figure(self, arguments):
        """Add figure to the report
        """

    @abc.abstractmethod
    def recursively_build_report(self, structure_tree):
        """Recursively navigate tree sructure to build report
        """

    @abc.abstractmethod
    def assemble(self, report_map, version=None, latex_proc=None):
        """Create a report possibly including assembler version
        """

        # Early exit if report map empty
        if not report_map:
            print("** ERROR: No content is available to populate report. Exiting")
            sys.exit(1)

        elif not "chapters" in report_map:
            print("** ERROR: Chapters are not defined in provided structure file. Exiting")
            sys.exit(1)

    def add_meta_information(self, item):
        """Add meta-information about data to the report
        """

        # Retrieve required parameters
        data_type = item["datatype"]
        data_set = item["dataset"]

        # Retrieve or assign optional verbosity parameter
        verbosity = self.VerbosityLevels.get("default")

        # Assemble data base name
        datafile = os.path.join(self.Parameters.DataDir, data_set)

        # Retrieve meta-information
        data = self.Parameters.DataFactory(
            data_type, datafile, item.get("parameters", {}))
        meta_info = data.get_meta_information()

        # Break out early if no meta-information was retrieved
        if not meta_info or not len(meta_info):
            print("*  WARNING: No meta-information for {} type available in {}".format(
                data_type, data_set))
            return

        # Generate meta-data for key/value files
        if data_type == "key-value":
            # Create table header with file name as first column
            name_str = "file name"
            other_cols = [k for k in meta_info[0] if k != name_str]
            header_list = [name_str] + other_cols

            # Create table body
            body_dict = {}
            for m in meta_info:
                key = m.get(name_str)

                # Ensure consistency of meta-information across accessors
                if not key or other_cols != [k for k in m if k != name_str]:
                    print("*  WARNING: Inconsistent meta-information for {}. Ignoring dataset {}".format(
                        data_type, data_set))
                    return

                # Update table body
                body_dict[key] = [m.get(c) for c in other_cols]

            caption_string = argMultiFontStringHelper(self)
            caption_string.append("Meta-information of ", "default")
            data_extension = item.get("parameters", {}).get("extension")
            if data_extension:
                caption_string.append(" all files with extension ", "default")
                caption_string.append(data_extension, "typewriter")
                caption_string.append(" in data directory.", "default")
            else:
                caption_string.append(data_set, "typewriter")
            caption_string.append(".", "default")
            self.add_table(
                header_list,
                body_dict,
                caption_string,
                True)

        # Generate meta-data for HDF5 file
        if data_type == "HDF5":
            # Iterate over per-file meta-information entries
            for m in meta_info:
                # Create caption string
                caption_string = argMultiFontStringHelper(self)
                caption_string.append("Meta-information of ", "default")
                caption_string.append(data_set, "typewriter")

                # Generate meta-information table
                self.add_table(
                    ["property", "value"],
                    {k: ["{}".format(v)] for k, v in m.items() if k != "file name"},
                    caption_string, False)

        # Generate meta-data for Exodus mesh
        elif data_type == "ExodusII":
            # Retrieve base name of file
            file_name = os.path.basename(datafile)

            # Eliminate final extension if file name is that of a partition
            if len(meta_info) > 1:
                file_name = os.path.splitext(file_name)[0]

            # Create table of topological properties
            header_list, body_dict = data.summarize(type="topology")
            caption_string = argMultiFontStringHelper(self)
            caption_string.append("Topological properties of ", "default")
            caption_string.append(file_name, "typewriter")
            self.add_table(header_list, body_dict, caption_string, True)

            # Create table of mesh blocks if requested and at least one exists
            if verbosity > self.VerbosityLevels.get("terse"):
                header_list, body_dict = data.summarize(type="blocks")
                if header_list:
                    caption_string.clear()
                    caption_string.append("Element blocks of ", "default")
                    caption_string.append(file_name, "typewriter")
                    self.add_table(header_list, body_dict, caption_string, True)

            # Create table of node and side sets if requested and at least one exists
            if verbosity > self.VerbosityLevels.get("terse"):
                for set_type in ("node", "side"):
                    header_list, body_dict = data.summarize(type="sets", set_type=set_type)
                    if header_list:
                        caption_string.clear()
                        caption_string.append(set_type.title(), "default")
                        caption_string.append(" sets of ", "default")
                        caption_string.append(file_name, "typewriter")
                        self.add_table(header_list, body_dict, caption_string, True)

            # Create table of variables if some are present
            header_list, body_list = data.summarize(type="variable")
            if body_list:
                caption_string.clear()
                caption_string.append("Variables of ", "default")
                caption_string.append(file_name, "typewriter")
                self.add_table(header_list, body_list, caption_string, True)


    def add_information(self, item):
        """Add information about given property to report and
           restrict information to specific items if provided
        """

        # Unpack property request and bail out early if empty
        property_request = item["property_request"]
        if not property_request:
            return

        # Retrieve request parameters
        data_type = item["datatype"]
        data_set = item["dataset"]

        # At least one data property was requested
        prop_type = property_request[0]
        prop_items = property_request[1:]

        # Assemble data base name and instantiate reader
        datafile = os.path.join(self.Parameters.DataDir, data_set)
        data = self.Parameters.DataFactory(
            data_type, datafile, item.get("parameters", {}))

        # Retrieve information for given property
        prop_info = data.get_property_information(prop_type, prop_items)

        # If no property values were found report it and break out
        if not prop_info:
            # Report missing values for property type
            multi_font_string = argMultiFontStringHelper(self)
            multi_font_string.append("No values for ", "default")
            p_items = False
            for p in prop_items:
                # Report missing values for property items
                if p_items:
                    multi_font_string.append(", ", "default")
                p_items = True
                multi_font_string.append(p, "typewriter")
            if p_items:
                multi_font_string.append(" items of ", "default")
            multi_font_string.append(prop_type, "typewriter")
            multi_font_string.append(" property were found.", "default")
            self.add_paragraph({"string": multi_font_string})

            # Do not add anything else when values are missing
            return

        # Some specific items were requested, build a summary table
        elif isinstance(prop_info, argInformationObject):
            # Create tables depending on information type
            prop_info_type = prop_info.get_type()

            # Handle dictionary of lists of values
            if prop_info_type == "arg_dict_lists_lists":
                # Create header common to all tables
                tab_head = []
                for p_name in prop_info.get_names():
                    multi_font_string = argMultiFontStringHelper(self)
                    multi_font_string.append(p_name, "typewriter")
                    tab_head.append(multi_font_string)

                # Iterate over property object items
                for info_key, info_value in prop_info.iterator():
                    multi_font_string = argMultiFontStringHelper(self)
                    if prop_type:
                        multi_font_string.append("Values of ", "default")
                        multi_font_string.append(prop_type, "typewriter")
                        multi_font_string.append(" property for ", "default")
                    else:
                        multi_font_string.append("Values for ", "default")
                    multi_font_string.append(info_key, "typewriter")
                    multi_font_string.append(".", "default")
                    self.add_table(
                        tab_head,
                        info_value,
                        multi_font_string,
                        True)

            # Unsupported type of information object
            else:
                print("*  WARNING: informtation type {} not supported by backend.".format(
                    prop_info_type))
                return

        # Some specific items were requested, build a summary table
        elif prop_items and isinstance(prop_info, dict):
            # Create table for property items only if needed
            tab_head = [argMultiFontStringHelper(self)]
            tab_head[0].append(prop_type, "typewriter")
            tab_head += [argMultiFontStringHelper(self)
                         for _ in prop_items]
            for p in prop_items:
                tab_head.append(p)
            multi_font_string = argMultiFontStringHelper(self)
            multi_font_string.append("Values of ", "default")
            multi_font_string.append(prop_type, "typewriter")
            multi_font_string.append(" properties.", "default")
            self.add_table(
                tab_head,
                dict(prop_info),
                multi_font_string,
                True)

        # No specific item required, build many tables
        else:
            # Iterate over of properties information list
            for props in prop_info:
                # Skip void entries
                if not props:
                    continue

                # Handle list-based properties
                if isinstance(props, list) and len(props) > 2:
                    prop_item = props[0]
                    prop_values = props[1]
                    if prop_item and prop_values and isinstance(prop_values, dict):
                        tab_head = [argMultiFontStringHelper(self),
                                    argMultiFontStringHelper(self)]
                        tab_head[0].append(prop_type, "typewriter")
                        tab_head[1].append(prop_item, "typewriter")
                        multi_font_string = argMultiFontStringHelper(self)
                        multi_font_string.append("Values of ", "default")
                        multi_font_string.append(prop_item, "typewriter")
                        multi_font_string.append(" for property ", "default")
                        multi_font_string.append(prop_type, "typewriter")
                        multi_font_string.append('.', "default")
                        self.add_table(
                            tab_head,
                            list(prop_values.items()),
                            multi_font_string,
                            True)


    def add_aggregation(self, item):
        """Add information aggregated from various backends
        """

        # Choose aggregator depending on data type
        data_type = item.get("datatype")
        if data_type == "ExodusII":
            # Instantiate ExodusII aggregator on self and execute it
            aggregator = argExodusAggregator.argExodusAggregator(self, item)
            aggregator.aggregate()
        elif data_type == "vtkSTL":
            # Instantiate vtkSNL aggregator on self and execute it
            aggregator = argVTKSTLAggregator.argVTKSTLAggregator(self, item)
            aggregator.aggregate()
        else:
            # Unsupported data type
            print("[argBackendBase] Unknown aggregation data type: {}")
            

    def fetch_image_and_caption(self, arguments):
        """Retrieve image and associated caption for figure creation
        """

        # A figure file is required
        figure_file = arguments.get("figure_file")
        if not figure_file:
            print("*  WARNING: no figure file name was provided.")
            return None, None

        # Determine path and file names
        figure_file_name = os.path.basename(figure_file)

        # Ensure that image file is present in output directory
        img_file = os.path.join(self.Parameters.OutputDir,
                                figure_file_name)
        if not os.path.isfile(img_file):
            # Ensure that destination directory exists
            if not os.path.exists(self.Parameters.OutputDir):
                os.makedirs(self.Parameters.OutputDir)

            # Try to copy image file from data directory
            try:
                shutil.copyfile(
                    os.path.join(self.Parameters.DataDir, figure_file),
                    img_file)
                # Could not copy picture file
            except IOError:
                print("*  WARNING: could find image file {} neither in {} nor {}. Ignoring figure.".format(
                    figure_file_name,
                    self.Parameters.DataDir,
                    self.Parameters.OutputDir))
                return None, None

        # Retrieve caption string when it is specified
        caption_string = arguments.get("caption_string")
        if not caption_string:
            # Otherwise try to retrieve caption from file
            caption_file = arguments.get("caption_file")
            if caption_file:
                # Try to fetch caption string first from data then output directories
                for d in (
                        self.Parameters.DataDir,
                        self.Parameters.OutputDir):
                    try:
                        with open(os.path.join(d, caption_file), 'r') as f:
                            caption_string = f.read()
                            break
                    except IOError:
                        continue
                else:  # for d in target_dirs
                    print("*  WARNING: could find caption file {} in neither {} nor {}.".format(
                        caption_file,
                        self.Parameters.DataDir,
                        self.Parameters.OutputDir))
                    caption_string = ''

        # Return retrieve image file name
        return figure_file_name, caption_string



#HEADER
#                        arg/GUI/Logic/argSaveAsAction.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

from PySide2.QtWidgets import QAction, QApplication, QFileDialog

app = "ARG-GUI"


class argSaveAsAction(QAction):

    def __init__(self):
        super().__init__()
        # scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        # self.setIcon(QIcon("{}/{}".format(scriptDirectory, "../Graphics/save.png")))
        self.setText("Save as...")
        self.setToolTip("Save currently displayed parameters value to specific location")

        self.triggered.connect(self.onTriggered)

    @staticmethod
    def onTriggered():
        """Routine when 'Save' action is triggered
        """

        # Log action
        print("[{}] 'Save as...' action detected.".format(app))

        # Retrieve settings
        settings = QApplication.instance().settingsController
        fileToSave = ""
        # fileToSave = QFileDialog.getSaveFileName(None,"Save Parameters File", "", "yaml file (*.yml *.yaml)")

        # Open FileDialog to get file information
        saveFileDialog = QFileDialog()
        saveFileDialog.setAcceptMode(QFileDialog.AcceptSave)
        saveFileDialog.setConfirmOverwrite(True)
        saveFileDialog.setOption(QFileDialog.DontConfirmOverwrite, False)
        saveFileDialog.setViewMode(QFileDialog.Detail)
        saveFileDialog.setNameFilter("yaml file (*.yml *.yaml)")
        saveFileDialog.setConfirmOverwrite(True)
        if saveFileDialog.exec_():
            fileNames = saveFileDialog.selectedFiles()
            if len(fileNames) == 1:
                fileToSave = fileNames[0]

                # Append with extension if missing
                if not (fileToSave.endswith(".yml") or fileToSave.endswith(".yaml")):
                    fileToSave = "{}.yml".format(fileToSave)

        # Log file name
        print("[{}] Saving to file \'{}\'".format(app, fileToSave))
        if fileToSave:
            QApplication.instance().saveRequested(fileToSave)



#HEADER
#                     arg/GUI/View/argUserSettingsDialog.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

from PySide2.QtCore import Qt, Slot
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QApplication, QGridLayout, QVBoxLayout, QLabel, QLineEdit, QSizePolicy, QSpacerItem, \
    QToolButton, QDialog, QDialogButtonBox, QMessageBox

from arg.GUI.Logic.argSettingsController import argSettingsController


class argUserSettingsDialog(QDialog):
    """A widget class to cover 'Insert' tab
    """

    def __init__(self, parent=None):
        super().__init__()
        self.setWindowTitle("ARG settings")

        self.SettingController = argSettingsController()

        # dialog Message box
        self.messageBox = QMessageBox()
        self.setModal(True)

        # Flags: disable minimize and close buttons
        self.setWindowFlags(Qt.Tool | Qt.WindowTitleHint | Qt.CustomizeWindowHint)

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        reloadIcon = QIcon("{}/{}".format(scriptDirectory, "../Graphics/reload.png"))
        helpIcon = QIcon("{}/{}".format(scriptDirectory, "../Graphics/help.png"))

        # Python exe path
        self.argPythonExePathLabel = QLabel(settings.getArgPythonExePathLabel())
        self.argPythonExePathLineEdit = QLineEdit()
        self.argPythonExePathReloadToolButton = QToolButton()
        self.argPythonExePathReloadToolButton.setIcon(reloadIcon)
        self.argPythonExePathHelpToolButton = QToolButton()
        self.argPythonExePathHelpToolButton.setIcon(helpIcon)

        # Python site package path
        self.argPythonSitePackageLabel = QLabel(settings.getArgPythonSitePackagePathLabel())
        self.argPythonSitePackageLineEdit = QLineEdit()
        self.argPythonSitePackageReloadToolButton = QToolButton()
        self.argPythonSitePackageReloadToolButton.setIcon(reloadIcon)
        self.argPythonSitePackageHelpToolButton = QToolButton()
        self.argPythonSitePackageHelpToolButton.setIcon(helpIcon)

        # ARG script path
        self.argScriptPathLabel = QLabel(settings.getArgScriptPathLabel())
        self.argScriptPathLineEdit = QLineEdit()
        self.argScriptPathReloadToolButton = QToolButton()
        self.argScriptPathReloadToolButton.setIcon(reloadIcon)
        self.argScriptPathHelpToolButton = QToolButton()
        self.argScriptPathHelpToolButton.setIcon(helpIcon)

        # Latex processor executable path
        self.argLatexProcessorPathLabel = QLabel(settings.getArgLatexProcessorPathLabel())
        self.argLatexProcessorPathLineEdit = QLineEdit()
        self.argLatexProcessorPathReloadToolButton = QToolButton()
        self.argLatexProcessorPathReloadToolButton.setIcon(reloadIcon)
        self.argLatexProcessorPathHelpToolButton = QToolButton()
        self.argLatexProcessorPathHelpToolButton.setIcon(helpIcon)

        # Instantiate the child layout and fill with widgets created above
        self.userSettingsChildLayout = QGridLayout()
        self.userSettingsChildLayout.setMargin(6)
        self.userSettingsChildLayout.setHorizontalSpacing(2)
        self.userSettingsChildLayout.addWidget(self.argPythonExePathLabel, 0, 0)
        self.userSettingsChildLayout.addWidget(self.argPythonExePathLineEdit, 0, 2)
        # self.userSettingsChildLayout.addWidget(self.argPythonExePathReloadToolButton, 0, 3)
        self.userSettingsChildLayout.addWidget(self.argPythonExePathHelpToolButton, 0, 4)
        self.userSettingsChildLayout.addWidget(self.argPythonSitePackageLabel, 1, 0)
        self.userSettingsChildLayout.addWidget(self.argPythonSitePackageLineEdit, 1, 2)
        # self.userSettingsChildLayout.addWidget(self.argPythonSitePackageReloadToolButton, 1, 3)
        self.userSettingsChildLayout.addWidget(self.argPythonSitePackageHelpToolButton, 1, 4)
        self.userSettingsChildLayout.addWidget(self.argScriptPathLabel, 2, 0)
        self.userSettingsChildLayout.addWidget(self.argScriptPathLineEdit, 2, 2)
        # self.userSettingsChildLayout.addWidget(self.argScriptPathReloadToolButton, 2, 3)
        self.userSettingsChildLayout.addWidget(self.argScriptPathHelpToolButton, 2, 4)
        self.userSettingsChildLayout.addWidget(self.argLatexProcessorPathLabel, 5, 0)
        self.userSettingsChildLayout.addWidget(self.argLatexProcessorPathLineEdit, 5, 2)
        # self.userSettingsChildLayout.addWidget(self.argLatexProcessorPathReloadToolButton, 5, 3)
        self.userSettingsChildLayout.addWidget(self.argLatexProcessorPathHelpToolButton, 5, 4)
        self.userSettingsChildLayout.setColumnMinimumWidth(1, 10)

        # Instantiate the main layout and fill with the child layout, the space and the button box
        self.userSettingsLayout = QVBoxLayout()
        self.userSettingsLayout.addLayout(self.userSettingsChildLayout)

        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        self.userSettingsLayout.addItem(QSpacerItem(0, 0, QSizePolicy.Ignored, QSizePolicy.Expanding))
        self.userSettingsLayout.addWidget(self.buttonBox)

        self.setLayout(self.userSettingsLayout)

        # Connections:
        # validation
        self.buttonBox.accepted.connect(self.onAccepted)

        # Help
        # self.argPythonExePathReloadToolButton.clicked.connect(self.onArgPythonExePathReloadToolButtonTriggered)
        self.argPythonExePathHelpToolButton.clicked.connect(self.onArgPythonExePathHelpToolButtonTriggred)
        # self.argPythonSitePackageReloadToolButton.clicked.connect(self.onArgPythonSitePackageReloadToolButtonTriggered)
        self.argPythonSitePackageHelpToolButton.clicked.connect(self.onArgPythonSitePackageHelpToolButtonTriggered)
        # self.argScriptPathReloadToolButton.clicked.connect(self.onArgScriptPathReloadToolButtonTriggered)
        self.argScriptPathHelpToolButton.clicked.connect(self.onArgScriptPathHelpToolButtonTriggered)
        # self.argLatexProcessorPathReloadToolButton.clicked.connect(self.onArgLatexProcessorPathReloadToolButtonTriggered)
        self.argLatexProcessorPathHelpToolButton.clicked.connect(self.onArgLatexProcessorPathHelpToolButtonTriggered)

    def initialize(self):
        """ Initialize user settings using the dialog lineedit values
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Initialize all LineEdits texts
        self.initiliazeAllLineEditTexts(settings)

        # Resize all LineEdits
        self.resizeAllLineEditsWidths(settings)

    def initiliazeAllLineEditTexts(self, settings):
        """ Initialize user settings using the dialog lineedit values
        """

        # Set all LineEdit texts
        self.argPythonExePathLineEdit.setText(settings.getArgPythonExePath())
        self.argPythonSitePackageLineEdit.setText(settings.getArgPythonSitePackagePath())
        self.argScriptPathLineEdit.setText(settings.getArgScriptPath())
        self.argLatexProcessorPathLineEdit.setText(settings.getArgLatexProcessorPath())

    def resizeAllLineEditsWidths(self, settings):
        """ Initialize user settings using the dialog lineedit values
        """

        # Instantiate list of LineEdit widths
        widths = []

        # Retrieve minimal wideth of each LineEdit as determined by font and text
        fontMetrics = self.argPythonExePathLineEdit.fontMetrics()
        widths.append(fontMetrics.boundingRect(settings.getArgPythonExePath()).width())
        fontMetrics = self.argPythonSitePackageLineEdit.fontMetrics()
        widths.append(fontMetrics.boundingRect(settings.getArgPythonSitePackagePath()).width())
        fontMetrics = self.argScriptPathLineEdit.fontMetrics()
        widths.append(fontMetrics.boundingRect(settings.getArgScriptPath()).width())
        fontMetrics = self.argLatexProcessorPathLineEdit.fontMetrics()
        widths.append(fontMetrics.boundingRect(settings.getArgLatexProcessorPath()).width())
        maxWidth = max(widths)

        # Resize all LineEdits to maximal required widths
        self.argPythonExePathLineEdit.resize(maxWidth, self.argPythonExePathLineEdit.height())
        self.argPythonSitePackageLineEdit.resize(maxWidth, self.argPythonSitePackageLineEdit.height())
        self.argScriptPathLineEdit.resize(maxWidth, self.argScriptPathLineEdit.height())
        self.argLatexProcessorPathLineEdit.resize(maxWidth, self.argLatexProcessorPathLineEdit.height())

        self.userSettingsChildLayout.setColumnMinimumWidth(2, maxWidth + self.userSettingsChildLayout.columnStretch(2))

    @Slot()
    def onArgPythonExePathReloadToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.argPythonExePathLineEdit.setText(settings.readConfigArgPythonExePath())

    @Slot()
    def onArgPythonExePathHelpToolButtonTriggred(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.warn(settings.getArgPythonExePathToolTip())

    @Slot()
    def onArgPythonSitePackageReloadToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.argPythonSitePackageLineEdit.setText(settings.readConfigArgPythonSitePackagePath())

    @Slot()
    def onArgPythonSitePackageHelpToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.warn(settings.getArgPythonSitePackagePathToolTip())

    @Slot()
    def onArgScriptPathReloadToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.argScriptPathLineEdit.setText(settings.readConfigArgScriptPath())

    @Slot()
    def onArgScriptPathHelpToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.warn(settings.getArgScriptPathToolTip())

    @Slot()
    def onArgLatexProcessorPathReloadToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.argLatexProcessorPathLineEdit.setText(settings.readConfigArgLatexProcessorPath())

    @Slot()
    def onArgLatexProcessorPathHelpToolButtonTriggered(self):
        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        self.warn(settings.getArgLatexProcessorPathToolTip())

    @Slot()
    def onAccepted(self):

        # Check current set of settings integrity
        settingsData = self.checkSettingsIntegrity()

        # Call settings controller to save current set of settings
        if settingsData:
            settings = QApplication.instance().settingsController
            settings.saveSettings(settingsData)

    def checkSettingsIntegrity(self):
        """Check current set of settings integrity
        """

        # Instantiate acceptance boolean
        accept = True

        # Instantiate settings data dictionary
        settingsData = {}

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        if self.argPythonExePathLineEdit.text() == "":
            accept = False
        else:
            settingsData[settings.getArgPythonExePathUserSettingsKey()] = self.argPythonExePathLineEdit.text()

        if self.argPythonSitePackageLineEdit.text() == "":
            accept = False
        else:
            settingsData[
                settings.getArgPythonSitePackagePathUserSettingsKey()] = self.argPythonSitePackageLineEdit.text()

        if self.argScriptPathLineEdit.text() == "":
            accept = False
        else:
            settingsData[settings.getArgScriptPathUserSettingsKey()] = self.argScriptPathLineEdit.text()

        if self.argLatexProcessorPathLineEdit.text() == "":
            accept = False
        else:
            settingsData[settings.getArgLatexProcessorPathUserSettingsKey()] = self.argLatexProcessorPathLineEdit.text()

        if accept == True:
            self.close()
        else:
            self.messageBox.setText("Please fill all user parameters to access ARG-GUI")
            self.messageBox.show()
            return

        # Return completed settings data dictionary
        return settingsData

    def warn(self, text):
        self.messageBox.setText(text)
        self.messageBox.show()



#HEADER
#                  arg/GUI/View/argReportInformationWidget.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os

from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QApplication, QFileDialog, QGridLayout, QGroupBox, QToolButton, QHBoxLayout, QLabel, \
    QLineEdit, QRadioButton, QSizePolicy, QSpacerItem, QWidget


class argReportInformationWidget(QWidget):
    """A widget class to cover 'Report Information' tab
    """

    def __init__(self):
        super().__init__()

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        icon = QIcon("{}/{}".format(scriptDirectory, "../Graphics/open.png"))

        # Set ARG execution default option
        self.argExecutionOption = settings.eOption()

        # Backend
        self.backendGroupBox = QGroupBox()
        self.backendLabel = QLabel(settings.backendTypeValue(settings.label))
        self.backendLatexRadioButton = QRadioButton(settings.backendTypeKeyLatex())
        self.backendWordRadioButton = QRadioButton(settings.backendTypeKeyWord())
        self.backendLayout = QHBoxLayout()
        self.backendLayout.addWidget(self.backendLatexRadioButton)
        self.backendLayout.addWidget(self.backendWordRadioButton)
        self.backendLayout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))
        self.backendGroupBox.setLayout(self.backendLayout)
        self.backendLatexRadioButton.setChecked(True)

        # Report Type
        self.reportTypeGroupBox = QGroupBox(self)
        self.reportTypeLabel = QLabel(settings.reportTypeValue(settings.label))
        self.reportTypeReportRadioButton = QRadioButton(settings.reportTypeKeyReport())
        self.reportTypeLayout = QHBoxLayout()
        self.reportTypeLayout.addWidget(self.reportTypeReportRadioButton)
        self.reportTypeLayout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))
        self.reportTypeGroupBox.setLayout(self.reportTypeLayout)
        self.reportTypeReportRadioButton.setChecked(True)

        # Classification
        self.classificationTypeGroupBox = QGroupBox(self)
        self.classificationTypeLabel = QLabel(settings.classificationValue(settings.label))
        self.classificationTypeLabel.setVisible(False)
        self.classificationTypeGenericRadioButton = QRadioButton(settings.classificationKeyGeneric())
        self.classificationTypeLayout = QHBoxLayout()
        self.classificationTypeLayout.addWidget(self.classificationTypeGenericRadioButton)
        self.classificationTypeLayout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))
        self.classificationTypeGroupBox.setLayout(self.classificationTypeLayout)
        self.classificationTypeGenericRadioButton.setChecked(True)
        self.classificationTypeGenericRadioButton.setVisible(False)
        self.classificationTypeGroupBox.setVisible(False)

        # File name
        self.fileNameLabel = QLabel(settings.fileNameValue(settings.label))
        self.fileNameLineEdit = QLineEdit()

        # Mutables File
        self.mutableFileLabel = QLabel(settings.mutablesFileValue(settings.label))
        self.mutableLineEdit = QLineEdit()
        self.mutableOpenButton = QToolButton()
        self.mutableOpenButton.setIcon(icon)
        self.mutableHLayout = QHBoxLayout()
        self.mutableHLayout.addWidget(self.mutableLineEdit)
        self.mutableHLayout.addWidget(self.mutableOpenButton)

        # Structure File
        self.structureFileLabel = QLabel(settings.structureFileValue(settings.label))
        self.structureFileLineEdit = QLineEdit()
        self.structureOpenButton = QToolButton()
        self.structureOpenButton.setIcon(icon)
        self.structureHLayout = QHBoxLayout()
        self.structureHLayout.addWidget(self.structureFileLineEdit)
        self.structureHLayout.addWidget(self.structureOpenButton)

        # Structure End File
        self.structureEndFileLabel = QLabel(settings.structureEndFileValue(settings.label))
        self.structureEndFileLineEdit = QLineEdit()
        self.structureEndOpenButton = QToolButton()
        self.structureEndOpenButton.setIcon(icon)
        self.structureEndHLayout = QHBoxLayout()
        self.structureEndHLayout.addWidget(self.structureEndFileLineEdit)
        self.structureEndHLayout.addWidget(self.structureEndOpenButton)

        # Artifact File
        self.artifactFileLabel = QLabel(settings.artifactFileValue(settings.label))
        self.artifactFileLineEdit = QLineEdit()
        self.artifactOpenButton = QToolButton()
        self.artifactOpenButton.setIcon(icon)
        self.artifactHLayout = QHBoxLayout()
        self.artifactHLayout.addWidget(self.artifactFileLineEdit)
        self.artifactHLayout.addWidget(self.artifactOpenButton)

        # Output Folder
        self.outputFolderLabel = QLabel(settings.outputFolderValue(settings.label))
        self.outputFolderLineEdit = QLineEdit()
        self.outputOpenButton = QToolButton()
        self.outputOpenButton.setIcon(icon)
        self.outputHLayout = QHBoxLayout()
        self.outputHLayout.addWidget(self.outputFolderLineEdit)
        self.outputHLayout.addWidget(self.outputOpenButton)

        # Verbosity
        self.verbosityLabel = QLabel(settings.verbosityValue(settings.label))
        self.verbosityGroupBox = QGroupBox(self)
        self.verbosityTerseRadioButton = QRadioButton(settings.verbosityKeyTerse())
        self.verbosityDefaultRadioButton = QRadioButton(settings.verbosityKeyDefault())
        self.verbosityVerboseRadioButton = QRadioButton(settings.verbosityKeyVerbose())
        self.verbosityDefaultRadioButton.setChecked(True)
        self.verbosityLayout = QHBoxLayout()
        self.verbosityLayout.addWidget(self.verbosityTerseRadioButton)
        self.verbosityLayout.addWidget(self.verbosityDefaultRadioButton)
        self.verbosityLayout.addWidget(self.verbosityVerboseRadioButton)
        self.verbosityLayout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))
        self.verbosityGroupBox.setLayout(self.verbosityLayout)

        # Instantiate 'Report Information' tab layout and fill with widgets created above
        self.reportInformationLayout = QGridLayout()
        self.reportInformationLayout.setMargin(5)
        self.reportInformationLayout.addWidget(self.backendLabel, 0, 0)
        self.reportInformationLayout.addWidget(self.backendGroupBox, 0, 2)
        self.reportInformationLayout.addWidget(self.reportTypeLabel, 1, 0)
        self.reportInformationLayout.addWidget(self.reportTypeGroupBox, 1, 2)
        self.reportInformationLayout.addWidget(self.classificationTypeLabel, 2, 0)
        self.reportInformationLayout.addWidget(self.classificationTypeGroupBox, 2, 2)
        self.reportInformationLayout.addWidget(self.fileNameLabel, 3, 0)
        self.reportInformationLayout.addWidget(self.fileNameLineEdit, 3, 2)
        self.reportInformationLayout.addWidget(self.mutableFileLabel, 4, 0)
        self.reportInformationLayout.addLayout(self.mutableHLayout, 4, 2)
        self.reportInformationLayout.addWidget(self.structureFileLabel, 5, 0)
        self.reportInformationLayout.addLayout(self.structureHLayout, 5, 2)
        self.reportInformationLayout.addWidget(self.structureEndFileLabel, 6, 0)
        self.reportInformationLayout.addLayout(self.structureEndHLayout, 6, 2)
        self.reportInformationLayout.addWidget(self.artifactFileLabel, 7, 0)
        self.reportInformationLayout.addLayout(self.artifactHLayout, 7, 2)
        self.reportInformationLayout.addWidget(self.outputFolderLabel, 8, 0)
        self.reportInformationLayout.addLayout(self.outputHLayout, 8, 2)
        self.reportInformationLayout.addWidget(self.verbosityLabel, 9, 0)
        self.reportInformationLayout.addWidget(self.verbosityGroupBox, 9, 2)
        self.reportInformationLayout.addItem(QSpacerItem(0, 0, QSizePolicy.Ignored, QSizePolicy.Expanding))
        self.reportInformationLayout.setColumnMinimumWidth(1, 5)
        self.setLayout(self.reportInformationLayout)

        # Connections
        self.reportTypeReportRadioButton.clicked.connect(self.onReportTypeChanged)

        self.mutableOpenButton.clicked.connect(self.onMutableOpenButtonTriggered)

        self.structureFileLineEdit.editingFinished.connect(self.onStructureFileLineEditEditingFinished)
        self.structureOpenButton.clicked.connect(self.onStructureOpenButtonTriggered)

        self.structureEndFileLineEdit.editingFinished.connect(self.onStructureEndFileLineEditEditingFinished)
        self.structureEndOpenButton.clicked.connect(self.onStructureEndOpenButtonTriggered)

        self.artifactFileLineEdit.editingFinished.connect(self.onArtifactFileLineEditEditingFinished)
        self.artifactOpenButton.clicked.connect(self.onArtifactOpenButtonTriggered)

        self.outputOpenButton.clicked.connect(self.onOutputOpenButtonTriggered)

    def fillParameters(self, data):
        """Set all 'Report Information' tab widgets to specified values
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Clean all widgets before filling them with appropriate values
        self.cleanParameters()

        # Update the display
        # Classification type
        if (settings.reportTypeKey() in data
                and data[settings.reportTypeKey()] == settings.reportTypeKeyReport()
                and settings.classificationKey() in data
                and data[settings.classificationKey()] != settings.classificationKeyGeneric()):
            classification = data[settings.classificationKey()]
            data[settings.classificationKey()] = settings.classificationKeyGeneric()
            print("*  WARNING: '{}' cannot have '{}' classification level. Setting to '{}'.".format(
                settings.reportTypeKeyReport(),
                classification,
                settings.classificationKeyGeneric()))

        # Report type
        value = data.get(settings.reportTypeKey())
        if value is None or value == settings.reportTypeKeyReport():
            self.displayClassification(False)
        else:
            self.displayClassification(True)

        # Backend
        value = data.get(settings.backendTypeKey())
        if value:
            if value == settings.backendTypeKeyLatex():
                self.backendLatexRadioButton.setChecked(True)
            elif value == settings.backendTypeKeyWord():
                self.backendWordRadioButton.setChecked(True)

        # Report type
        value = data.get(settings.reportTypeKey())
        if value == settings.reportTypeKeyReport():
            self.reportTypeReportRadioButton.setChecked(True)

        # Classification
        value = data.get(settings.classificationKey())
        if value == settings.classificationKeyGeneric():
            self.classificationTypeGenericRadioButton.setChecked(True)

        # File Name
        if data.get(settings.fileNameKey()):
            self.fileNameLineEdit.setText(data.get(settings.fileNameKey()))

        # Mutables File
        if data.get(settings.mutablesFileKey()):
            self.mutableLineEdit.setText(data.get(settings.mutablesFileKey()))

        # Structure File
        if data.get(settings.structureFileKey()):
            self.structureFileLineEdit.setText(data.get(settings.structureFileKey()))

        # Structure End File
        if data.get(settings.structureEndFileKey()):
            self.structureEndFileLineEdit.setText(data.get(settings.structureEndFileKey()))

        # Artifact File
        if data.get(settings.artifactFileKey()):
            self.artifactFileLineEdit.setText(data.get(settings.artifactFileKey()))

        # Output Folder
        if data.get(settings.outputFolderKey()):
            self.outputFolderLineEdit.setText(data.get(settings.outputFolderKey()))

        # Verbosity
        value = data.get(settings.verbosityKey())
        if value == settings.verbosityKeyTerseAsInt():
            self.verbosityTerseRadioButton.setChecked(True)
        elif value == settings.verbosityKeyDefaultAsInt():
            self.verbosityDefaultRadioButton.setChecked(True)
        elif value == settings.verbosityKeyVerboseAsInt():
            self.verbosityVerboseRadioButton.setChecked(True)

    def constructParameters(self):
        """Instantiate all 'Report Information' tab widgets and set to default
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Initialize parameters dict
        parameters = {}

        # Backend
        if self.backendLatexRadioButton.isChecked():
            parameters[settings.backendTypeKey()] = settings.backendTypeKeyLatex()
        elif self.backendWordRadioButton.isChecked():
            parameters[settings.backendTypeKey()] = settings.backendTypeKeyWord()

        # Report Type
        if self.reportTypeReportRadioButton.isChecked():
            parameters[settings.reportTypeKey()] = settings.reportTypeKeyReport()

        # Classification
        if self.classificationTypeGenericRadioButton.isChecked():
            parameters[settings.classificationKey()] = settings.classificationKeyGeneric().lower()

        # File Name
        parameters[settings.fileNameKey()] = self.fileNameLineEdit.text()

        # Mutables File
        parameters[settings.mutablesFileKey()] = self.mutableLineEdit.text()

        # Structure File
        parameters[settings.structureFileKey()] = self.structureFileLineEdit.text()

        # Structure End File
        parameters[settings.structureEndFileKey()] = self.structureEndFileLineEdit.text()

        # Artifact File
        parameters[settings.artifactFileKey()] = self.artifactFileLineEdit.text()

        # Output Folder
        parameters[settings.outputFolderKey()] = self.outputFolderLineEdit.text()

        # Verbosity
        if self.verbosityTerseRadioButton.isChecked():
            parameters[settings.verbosityKey()] = settings.verbosityKeyTerseAsInt()
        elif self.verbosityDefaultRadioButton.isChecked():
            parameters[settings.verbosityKey()] = settings.verbosityKeyDefaultAsInt()
        elif self.verbosityVerboseRadioButton.isChecked():
            parameters[settings.verbosityKey()] = settings.verbosityKeyVerboseAsInt()

        return parameters

    def cleanParameters(self):
        """Reset all 'Report Information' tab widgets to default
        """

        self.backendLatexRadioButton.setChecked(True)
        self.reportTypeReportRadioButton.setChecked(True)
        self.classificationTypeGenericRadioButton.setChecked(True)
        self.mutableLineEdit.setText("")
        self.structureFileLineEdit.setText("")
        self.structureEndFileLineEdit.setText("")
        self.artifactFileLineEdit.setText("")
        self.outputFolderLineEdit.setText("")
        self.verbosityDefaultRadioButton.setChecked(True)

    @staticmethod
    def onFillLineEdit(lineEdit, fileMode):
        openFileDialog = QFileDialog()
        openFileDialog.setAcceptMode(QFileDialog.AcceptOpen)
        openFileDialog.setFileMode(fileMode)
        openFileDialog.setViewMode(QFileDialog.Detail)
        if openFileDialog.exec_():
            fileNames = openFileDialog.selectedFiles()
            if len(fileNames) == 1:
                lineEdit.setText(fileNames[0])

    def onMutableOpenButtonTriggered(self):
        self.onFillLineEdit(self.mutableLineEdit, QFileDialog.AnyFile)

    def onStructureOpenButtonTriggered(self):
        self.onFillLineEdit(self.structureFileLineEdit, QFileDialog.AnyFile)
        self.onStructureFileLineEditEditingFinished()

    def onStructureEndOpenButtonTriggered(self):
        self.onFillLineEdit(self.structureEndFileLineEdit, QFileDialog.AnyFile)
        self.onStructureEndFileLineEditEditingFinished()

    def onArtifactOpenButtonTriggered(self):
        self.onFillLineEdit(self.artifactFileLineEdit, QFileDialog.AnyFile)
        self.onArtifactFileLineEditEditingFinished()

    def onOutputOpenButtonTriggered(self):
        self.onFillLineEdit(self.outputFolderLineEdit, QFileDialog.DirectoryOnly)

    def displayClassification(self, display):
        self.classificationTypeLabel.setVisible(display)
        self.classificationTypeGroupBox.setVisible(display)

    def onReportTypeChanged(self):
        if self.reportTypeReportRadioButton.isChecked():
            self.classificationTypeLabel.setVisible(False)
            self.classificationTypeGroupBox.setVisible(False)
            self.classificationTypeGenericRadioButton.setChecked(True)
        else:
            self.classificationTypeLabel.setVisible(True)
            self.classificationTypeGroupBox.setVisible(True)
            self.classificationTypeGenericRadioButton.setChecked(False)

    def onStructureFileLineEditEditingFinished(self):
        settings = QApplication.instance().settingsController

        if self.argExecutionOption != settings.eOption():
            QApplication.instance().checkLineEdit(self.structureFileLineEdit, False)

    def onStructureEndFileLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.structureEndFileLineEdit, False)

    def onArtifactFileLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.artifactFileLineEdit, False)

    def runEButtonClicked(self):
        """ Switch to E execution mode
        """

        settings = QApplication.instance().settingsController
        self.argExecutionOption = settings.eOption()
        self.structureFileLineEdit.setStyleSheet(QApplication.instance().getStyleSheetLineEditNormal())

    def runGButtonClicked(self):
        """ Switch to G execution mode
        """

        settings = QApplication.instance().settingsController
        self.argExecutionOption = settings.gOption()
        QApplication.instance().checkLineEdit(self.structureFileLineEdit, False)



# HEADER
#                     arg/Aggregation/argVTKSTLAggregator.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import os
import re

from arg.Common.argMultiFontStringHelper import argMultiFontStringHelper
from arg.DataInterface.argDataInterface import argDataInterface
from arg.Aggregation.argAggregatorBase import argAggregatorBase


class argVTKSTLAggregator(argAggregatorBase):
    """A class to aggregate information in a Exodus-specific way
    """

    def __init__(self, b, r):

        # Call superclass init
        super().__init__(b, r)


    def show_CAD_metadata(self):
        """Aggregate and show CAD metadata information
        """

        # Get handle on metadata and bail out early if absent
        metadata = self.RequestParameters.get("metadata")
        if not metadata:
            print("*  WARNING: igoring request: no metadata to aggregate")
            return

        # Instantiate interface to CAD properties data
        parameters_root = self.RequestParameters.get("parameters_root")
        regexp = re.compile(r"(.*)_parameters\.txt")
        cad_metadata = argDataInterface.factory(
            "key-value",
            os.path.join(self.Backend.Parameters.DataDir, parameters_root),
            {"expression": regexp})

        # Initialize body_lists
        all_body_lists = {}

        # Iterate over metadata items
        for metadatum in metadata:
            # Get information for current metadatum and iterate over it
            prop_info = cad_metadata.get_property_information(metadatum)
            for info_key, info_value in prop_info.iterator():
                # Initialize default values
                value = info_value[0][0]
                color, style = None, "default"

                # Handle particular values
                if value in ('', '-'):
                    color, value = "orange", "UNDEFINED"
                elif not value:
                    color, value = "red", "NOT FOUND"
                else:
                    style = "typewriter"

                # Update body list for this parameter file and value
                row = [argMultiFontStringHelper(self.Backend),
                       argMultiFontStringHelper(self.Backend)]
                row[0].append(prop_info.get_names()[0], "typewriter")
                row[1].append("{}".format(value), style, color)
                all_body_lists.setdefault(info_key, []).append(row)

        # Create one table per file
        for file_name, body_list in all_body_lists.items():
            self.Backend.add_table(
                ["CAD parameter", "parameter value"],
                body_list,
                "CAD metadata for part {}. ".format(
                    file_name.replace("_parameters.txt", '')))


    def aggregate(self):
        """Decide which aggregation operation is to be performed
        """

        # Switch between different aggregation types
        try:
            request_name = self.RequestParameters["name"]
        except:
            print("*  WARNING: ignoring request: no aggregation methodn name")
        print("[argVTKSTLAggregator] Processing {} request".format(request_name))

        # Operation show_CAD_metadata: create one table per reported CAD metadata
        if request_name == "show_CAD_metadata":
            self.show_CAD_metadata()

        # Operation show_mesh_surface: create one figure for the entire mesh
        elif request_name.startswith("show_mesh_surface"):
            # Decide whether mesh edges are to be shown or not
            self.RequestParameters["edges"] = request_name.endswith("with_edges")

            # Add specific request parameters
            self.RequestParameters.setdefault("view_direction", ())

            # Get handle on data
            file_name = self.RequestParameters["model"]
            data = argDataInterface.factory(
                "vtkSTL",
                os.path.join(self.Backend.Parameters.DataDir, file_name),
                self.RequestParameters.get("merge", "True"))

            # Aggregate
            if data:
                self.show_mesh_surface(data, file_name)



#HEADER
#                        arg/Generation/argMath.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

def aggregate_histograms(h_1, h_2):
    """Aggregate two histograms
       It is assumed that second histogram is a non-void dict
    """

    # Nothing to aggregate if first histogram is empty
    if not h_1:
        return dict(h_2)

    # Build set union of histogram keys
    s_keys = set(h_1) | set(h_2)

    # Return aggregated histogram
    return {
        k: h_1.get(k, 0) + h_2.get(k, 0) for k in s_keys}


def aggregate_descriptive_statistics(s_1, s_2):
    """Aggregate two sets of descriptive statistics (min/mean/max/M2/card)
       It is assumed that second set is an initialized 5-vector
    """

    # Nothing to aggregate if first set of statistics is empty
    if not s_1:
        return s_2[:]

    # Compute global cardinality and its inverse
    n_1 = s_1[4]
    n_2 = s_2[4]
    n_tot = n_1 + n_2
    inv_n_tot = 1. / n_tot

    # Compute means difference
    m_1 = s_1[1]
    m_2 = s_2[1]
    d_21 = m_2 - m_1

    # Retrieve M2 aggregates and compute global one
    M2_1 = s_1[3]
    if n_1 > 1:
        # Assume that unbiased estimator is used
        M2_1 *= (n_1 - 1)
    M2_2 = s_2[3]
    if n_2 > 1:
        # Assume that unbiased estimator is used
        M2_2 *= (n_2 - 1)
    M2 = M2_1 + M2_2 + n_1 * n_2 * d_21 * d_21 * inv_n_tot

    # Return aggregated statistics
    return [
        min(s_1[0], s_2[0]),
        m_1 + n_2 * d_21 * inv_n_tot,
        max(s_1[2], s_2[2]),
        M2 / (n_tot - 1) if n_tot > 1 else M2,
        n_tot]
